import re

CYRILLIC_SYMBOLS = 'АВЕСКМОТРХУаескопртху'
LATIN_SYMBOLS = 'ABECKMOTPXYaeckonpmxy'

EN_VOWELS = 'aoeiyu'
EE_VOWELS = 'aäoeöõiüu'
SE_VOWELS = 'aåäeöiyou'
FI_VOWELS = 'aäoeöiyu'
RU_VOWELS = 'аяоыиеёэюу'
RU_DEAF_VOWELS = 'аоыиэу'

EE_FI_DEAF_CONSONANTS = 'bvdg'
EN_DEAF_CONSONANTS = 'dgqvz'
SE_DEAF_CONSONANTS = 'bvdg'
RU_DEAF_CONSONANTS = 'бздвг'
RU_CONSONANTS = 'бвгджзклмнпрстфхцчшщ'

SE_VOWL = 'eiyöäj'

J_SEQ_RU = r'^|ъ|ь'
J_VOWEL_SEQ_RU = r'|'.join(RU_DEAF_VOWELS)
CONS_VOWEL_SEQ_RU = r'|'.join(RU_CONSONANTS)

RU_EGO_OGO_ENDING = re.compile(r'([ео])(г)(о$)', re.I)
RU_IA_ENDING = re.compile(r'[еи][ая]', re.I)
RU_II_ENDING = re.compile(r'и[еио]', re.I)

EN_REMOVE_MAP = [
    (re.compile(r'[hw]', re.I), '')
]

EN_PHONEMES = [
    (re.compile(r'^ye', re.I), 'je'),
    (re.compile(r'au', re.I), 'o'),
    (re.compile(r'e[ae]', re.I), 'e'),
    (re.compile(r'(ew|ue|oo)', re.I), 'u'),
    (re.compile(r'([ie]y|ai)', re.I), 'ei'),
    (re.compile(r'([iy]e|[iy]re)$', re.I), 'ai'),
    (re.compile(r'(e[ae]r|ere)$', re.I), 'ie')
]

EN_METAPHONE_PHONEMES = [
    (re.compile(r'^([kgp])(n)', re.I), r'\2'),
    (re.compile(r'^ae', re.I), 'e'),
    (re.compile(r'^wr', re.I), 'r'),
    (re.compile(r'mb$', re.I), 'm'),
    (re.compile(r'(c)(ia)', re.I), r'x\2'),
    (re.compile(r'sch', re.I), 'skh'),
    (re.compile(r'[cs]h', re.I), 'x'),
    (re.compile(r'th', re.I), '0'),
    (re.compile(r'(c)([iey])', re.I), r's\2'),
    (re.compile(r'c', re.I), 'k'),
    (re.compile(r'(d)(g[iey])', re.I), r'j\2'),
    (re.compile(rf'(g)(h)(?!($|[{EN_VOWELS}]))', re.I), r'j\2'),
    (re.compile(r'(g)(n|ned)$', re.I), r'j\2'),
    (re.compile(r'(?<!g)(g)([iey])', re.I), r'j\2'),
    (re.compile(rf'(?<![{EN_VOWELS}])(h)([{EN_VOWELS}])', re.I), r'\2'),
    (re.compile(r'ck', re.I), 'k'),
    (re.compile(r'ph', re.I), 'f'),
    (re.compile(r'^wh', re.I), 'w'),
    (re.compile(rf'(w)([^{EN_VOWELS}])', re.I), r'\2')
]

EN_VOWELS_TO_REMOVE = [(re.compile(rf'[{EN_VOWELS}]', re.I), '')]

EE_PHONEMES = [
    (re.compile(r'[cz]', re.I), 'ts'),
    (re.compile(r'q', re.I), 'kv'),
    (re.compile(r'w', re.I), 'v'),
    (re.compile(r'x', re.I), 'ks'),
    (re.compile(r'y', re.I), 'i')
]

FI_PHONEMES = [
    (re.compile(r'sh', re.I), 's'),
    (re.compile(r'ng', re.I), 'n'),
    (re.compile(r'z', re.I), 'ts'),
    (re.compile(r'q', re.I), 'kv'),
    (re.compile(r'w', re.I), 'v'),
    (re.compile(r'x', re.I), 'ks')
]

RU_PHONEMES = [
    (re.compile(r'(ст|[сзж])(ч)', re.I), r'щ'),  # Подписчик, заказчик, мужчина, жёстче.
    (re.compile(r'([сжт])(ш)', re.I), r'\2'),  # Обветшалый, выросший.
    (re.compile(r'([зс])(ж)', re.I), r'\2'),  # Визжать, обезжиренный.
    (re.compile(r'(с)(щ)', re.I), r'\2'),  # Бесщадный, расщепить.
    (re.compile(r'(т)([чщ])', re.I), r'ч'),  # Тщательно.
    (re.compile(r'(с)(т)([глнц])', re.I), r'\1\3'),  # Устлан, местный, бюстгальтер.
    (re.compile(r'(н)([тд])(ств)', re.I), r'\1\3'),  # Агентство, президентство.
    (re.compile(r'([нс])([тдк])(ск)', re.I), r'\1\3'),  # Баскский.
    (re.compile(r'(р)([гк])(ск)', re.I), r'\1\3'),  # Выборгский.
    (re.compile(r'(р)(д)([чц])', re.I), r'\1\3'),  # Сердце.
    (re.compile(r'(з)(д)([нц])', re.I), r'\1\3'),  # Уздцы, праздный, поздно.
    (re.compile(r'(ль|н)(д)([шц])', re.I), r'\1\3'),  # Голландцы, эндшпиль.
    (re.compile(r'(н)(т)(г)', re.I), r'\1\3'),  # Рентген.
    (re.compile(r'(в)(ств)', re.I), r'\2'),  # Здравствуй, чувствовать.
    (re.compile(r'(л)(нц)', re.I), r'\2'),  # Солнце.
    (re.compile(r'([дт][сц])', re.I), 'ц'),  # Солдатский, детство.
    (re.compile(f'(х)(г)', re.I), r'\2')  # Бухгалтер.
]

SE_PHONEMES = [
    (re.compile(r'^och$', re.I), 'å'),
    (re.compile(r'(rs|sch|ssj)', re.I), 'sh'),
    (re.compile(r'(stj|skj|sj|ch)', re.I), 'hf'),
    (re.compile(rf'(sk)([{SE_VOWL}])', re.I), r'sh\2'),
    (re.compile(rf'(k)([{SE_VOWL}])', re.I), r'sh\2'),
    (re.compile(rf'(c)([{SE_VOWL}])', re.I), r's\2'),
    (re.compile(rf'(g)([{SE_VOWL}])', re.I), r'j\2'),
    (re.compile(r'([dghl])(j)', re.I), r'\2'),
    (re.compile(r'(r)([ntl])', re.I), r'\2'),
    (re.compile(r'(i)(g)($)', re.I), r'\1\3'),
    (re.compile(r'(n)(g)', re.I), r'\1'),
    (re.compile(r'([cq]|ck)', re.I), 'k'),
    (re.compile(r'[st]ion', re.I), 'shn'),
    (re.compile(r'w', re.I), 'v'),
    (re.compile(r'x', re.I), 'ks'),
    (re.compile(r'z', re.I), 's'),
    (re.compile(r'tj', re.I), 'sh')
]

RU_REPLACEMENT_J_MAP = [
    (re.compile(r'(' + J_SEQ_RU + r')(я)', re.I), 'jа'),
    (re.compile(r'(' + J_SEQ_RU + r')(ю)', re.I), 'jу'),
    (re.compile(r'(' + J_SEQ_RU + r')(е)', re.I), 'jэ'),
    (re.compile(r'(' + J_SEQ_RU + r')(ё)', re.I), 'jо')
]

RU_REPLACEMENT_VOWEL_MAP = [
    (re.compile(r'(' + J_VOWEL_SEQ_RU + r')(я)', re.I), r'\1jа'),
    (re.compile(r'(' + J_VOWEL_SEQ_RU + r')(ю)', re.I), r'\1jу'),
    (re.compile(r'(' + J_VOWEL_SEQ_RU + r')(е)', re.I), r'\1jэ'),
    (re.compile(r'(' + J_VOWEL_SEQ_RU + r')(ё)', re.I), r'\1jо')
]

RU_REPLACEMENT_CONSONANT_MAP = [
    (re.compile(r'(' + CONS_VOWEL_SEQ_RU + r')(я)', re.I), r'\1а'),
    (re.compile(r'(' + CONS_VOWEL_SEQ_RU + r')(ю)', re.I), r'\1у'),
    (re.compile(r'(' + CONS_VOWEL_SEQ_RU + r')(е)', re.I), r'\1э'),
    (re.compile(r'(' + CONS_VOWEL_SEQ_RU + r')(ё)', re.I), r'\1о')
]

RU_REMOVE_MAP = [
    (re.compile(r'й', re.I), 'j'),
    (re.compile(r'[ъь]', re.I), '')
]

RU_VOWELS_TO_REMOVE = [
    (re.compile(rf'[{RU_VOWELS}]', re.I), ''),
    (re.compile(r'j', re.I), '')
]



from .base.base import BasePhoneticsAlgorithm
from .config import FI_VOWELS, RU_VOWELS, EE_VOWELS, RU_DEAF_CONSONANTS, \
    EE_FI_DEAF_CONSONANTS, SE_VOWELS, SE_DEAF_CONSONANTS, EN_DEAF_CONSONANTS, EN_VOWELS
from .ruleset import EstonianRuleSet, FinnishRuleSet, RussianRuleSet, SwedenRuleSet, EnglishMetaphoneRuleSet


class Metaphone(BasePhoneticsAlgorithm):
    """
    Basic class for Metaphone algorithm
    """
    def __init__(self, compress_ending=False, reduce_word=True):
        """
        Initialization of Metaphone object
        :param compress_ending: not used
        :param reduce_word: remove repeated letters from word
        """
        self._compress_ending = compress_ending
        self._reduce_word = reduce_word

    _deaf_consonants_seq = ''
    _deaf_consonants = str.maketrans(_deaf_consonants_seq, '')
    _vowels_table = str.maketrans('', '')

    def _deaf_consonants_letters(self, word):
        return word

    def _reduce_deaf_consonants_letters(self, word, criteria):
        res = []
        for i, letter in enumerate(word):
            if letter in self._deaf_consonants_seq and (i == len(word) - 1 or word[i + 1].lower() not in criteria):
                letter = letter.translate(self._deaf_consonants)
            res += [letter]
        return ''.join(res)

    @staticmethod
    def __compress_word_ending(word):
        return word

    def _apply_metaphone_algorithm(self, word):
        if self._reduce_word:
            word = self._reduce_seq(word)
        word = word.translate(self._vowels_table)
        word = self._deaf_consonants_letters(word)
        if self._compress_ending:
            word = self.__compress_word_ending(word)
        return word.upper()

    def transform(self, word):
        return self._apply_metaphone_algorithm(word)


class EnglishMetaphone(Metaphone):
    """
    Metaphone for English language
    """
    __rule_set = EnglishMetaphoneRuleSet()

    _vowels = EN_VOWELS
    _vowels_table = str.maketrans(EN_VOWELS, EN_VOWELS)
    _deaf_consonants_seq = EN_DEAF_CONSONANTS
    _deaf_consonants = str.maketrans(_deaf_consonants_seq, 'tkkfs')

    def __init__(self, compress_ending=False, reduce_word=True, reduce_vowels=False):
        """
        Initialization of Russian Metaphone object
        :param compress_ending: compress an ending of the transcribed word
        :param reduce_word: simplify sequences, remove duplicated adjacent letters
        :param reduce_vowels: simplify sequences of English vowels
        """
        super().__init__(compress_ending, reduce_word)
        self.reduce_vowels = reduce_vowels

    def _deaf_consonants_letters(self, word):
        return self._reduce_deaf_consonants_letters(word, self._vowels)

    def transform(self, word):
        word = self._cyrillic2latin(word)
        word = self.__rule_set.reduce_phonemes(word)
        word = self._apply_metaphone_algorithm(word)
        if self.reduce_vowels:
            word = word[0] + self.__rule_set.reduce_vowels(word[1:])
        return word


class RussianMetaphone(Metaphone):
    """
    Metaphone for Russian language
    """

    _vowels = RU_VOWELS
    _deaf_consonants_seq = RU_DEAF_CONSONANTS
    _deaf_consonants = str.maketrans(_deaf_consonants_seq, 'пстфк')
    _vowels_table = str.maketrans(_vowels, 'ААААИИИИУУ')

    def __init__(self, compress_ending=False, reduce_word=True, reduce_phonemes=False,
                 replace_ego_ogo_endings=False, deaf_all_consonants=False, reduce_vowels=False):
        """
        Initialization of Russian Metaphone object
        :param compress_ending: compress an ending of the transcribed word
        :param reduce_word: simplify sequences, remove duplicated adjacent letters
        :param reduce_phonemes: simplify sequences of Russian consonants
        :param replace_ego_ogo_endings: replace "-его/-ого" endings with "-ево/-ово"
        :param deaf_all_consonants: deaf consonants before Russian vowels
        :param reduce_vowels: delete all vowels exclude if a vowel if the first letter
        """
        super().__init__(compress_ending, reduce_word)
        self.reduce_phonemes = reduce_phonemes
        self.replace_ego_ogo_endings = replace_ego_ogo_endings
        self.deaf_all_consonants = deaf_all_consonants
        self.reduce_vowels = reduce_vowels
        self.rule_set = RussianRuleSet()

    def __replace_j_vowels(self, word):
        word = self.rule_set.replace_consonant_vowels(word)
        word = self.rule_set.replace_j_vowel_phonemes(word)
        word = self.rule_set.replace_j_and_signs(word)
        return self.rule_set.replace_ii_ending(word)

    @staticmethod
    def _compress_ending(word):
        return word

    def _deaf_consonants_letters(self, word):
        criterion = '' if self.deaf_all_consonants else 'лмнр' + self._vowels
        return self._reduce_deaf_consonants_letters(word, criterion)

    def transform(self, word):
        word = self._latin2cyrillic(word)
        if self.replace_ego_ogo_endings:
            word = self.rule_set.replace_ego_ogo_ending(word)
        if self.reduce_phonemes:
            word = self.rule_set.reduce_phonemes(word)
        word = self.__replace_j_vowels(word)
        word = self._apply_metaphone_algorithm(word)
        if self.reduce_vowels:
            fitsr_letter_idx = 2 if word.startswith('J') else 1
            word = word[:fitsr_letter_idx] + self.rule_set.reduce_vowels(word[fitsr_letter_idx:])
        return word


class FinnishMetaphone(Metaphone):
    """
    Metaphone for Finnish language
    """
    __rule_set = FinnishRuleSet()

    _vowels = FI_VOWELS
    _deaf_consonants_seq = EE_FI_DEAF_CONSONANTS
    _deaf_consonants = str.maketrans(_deaf_consonants_seq, 'pftk')
    _vowels_table = str.maketrans(FI_VOWELS, 'AAAIIIUU')

    def _deaf_consonants_letters(self, word):
        return self._reduce_deaf_consonants_letters(word, self._vowels + 'lmnr')

    def transform(self, word):
        word = self._cyrillic2latin(word)
        word = self.__rule_set.reduce_phonemes(word)
        return self._apply_metaphone_algorithm(word)


class EstonianMetaphone(Metaphone):
    """
    Metaphone for Estonian language
    """
    __rule_set = EstonianRuleSet()

    _vowels = EE_VOWELS
    _deaf_consonants_seq = EE_FI_DEAF_CONSONANTS
    _deaf_consonants = str.maketrans(_deaf_consonants_seq, 'pftk')
    _vowels_table = str.maketrans(EE_VOWELS, 'AAAIIIIUU')

    def _deaf_consonants_letters(self, word):
        return self._reduce_deaf_consonants_letters(word, self._vowels + 'lmnr')

    def transform(self, word):
        word = self._cyrillic2latin(word)
        word = self.__rule_set.reduce_phonemes(word)
        return self._apply_metaphone_algorithm(word)


class SwedenMetaphone(Metaphone):
    """
    Metaphone for Sweden language
    """
    __rule_set = SwedenRuleSet()

    _vowels = SE_VOWELS
    _deaf_consonants_seq = SE_DEAF_CONSONANTS
    _deaf_consonants = str.maketrans(_deaf_consonants_seq, 'pftk')
    _vowels_table = str.maketrans(SE_VOWELS, 'AAIIIIIUU')

    def _deaf_consonants_letters(self, word):
        return self._reduce_deaf_consonants_letters(word, self._vowels + 'lmnr')

    def transform(self, word):
        word = self._cyrillic2latin(word)
        if word.endswith('on') and not word.endswith('hon'):
            word = word[:-2] + 'ån'
        word = self.__rule_set.reduce_phonemes(word)
        return self._apply_metaphone_algorithm(word)



import editdistance

from abc import abstractmethod

from .base.base import BasePhoneticsAlgorithm
from .metaphone import Metaphone
from .soundex import Soundex


class PhoneticDistanceException(Exception):
    def __init__(self, msg):
        self.msg = msg


class PhoneticsDistance:
    @staticmethod
    def _hamming(word1, word2):
        if len(word1) != len(word2):
            raise PhoneticDistanceException('For Hamming distance words should be the same length!')
        return sum(a != b for a, b in zip(word1, word2))

    @staticmethod
    def _levenstein(word1, word2):
        return editdistance.eval(word1, word2)

    _distance_metric = {
        'levenstein': _levenstein.__func__,
        'hamming': _hamming.__func__,
    }

    @abstractmethod
    def distance(self, word1, word2):
        """
        Compute the distance between phonetics codes
        :param word1: first original word
        :param word2: second original word
        :return: distance value
        """
        return None


class PhoneticsInnerLanguageDistance(PhoneticsDistance):
    def __init__(self, phonetics, metric_name='levenstein', metrics=None):
        """
        Init a distance object
        :param phonetics: an object of BasePhoneticsAlgorithm class
        :param metric_name: distance function name, optional, default is Levenstein distance
        :param metrics: another distance function, optional
        """
        assert isinstance(phonetics, BasePhoneticsAlgorithm)
        assert metric_name in self._distance_metric.keys()
        self.phonetics = phonetics

        if not metrics:
            self.metrics = self._distance_metric[metric_name]
        else:
            self.metrics = metrics

    def distance(self, word1, word2):
        """
        Compute the distance between phonetics codes
        :param word1: first original word
        :param word2: second original word
        :return: distance value
        """
        w1, w2 = self.phonetics.transform(word1), self.phonetics.transform(word2)
        if isinstance(self.phonetics, Soundex):
            return self.metrics(w1, w2) if self.phonetics.is_delete_first_letter() else self.metrics(w1[1:], w2[1:])
        return self.metrics(w1, w2)


class PhoneticsBetweenLanguagesDistance(PhoneticsDistance):
    def __init__(self, phonetics1, phonetics2, metric_name='levenstein', metrics=None):
        """
        Init a distance object
        :param phonetics1: first object of BasePhoneticsAlgorithm class
        :param phonetics2: second object of BasePhoneticsAlgorithm class
        :param metric_name: distance function name, optional, default is Levenstein distance
        :param metrics: another distance function, optional
        """
        assert (isinstance(phonetics1, Soundex) and isinstance(phonetics2, Soundex) or
                isinstance(phonetics1, Metaphone) and isinstance(phonetics2, Metaphone))
        assert metric_name in self._distance_metric.keys()
        self.phonetics1 = phonetics1
        self.phonetics2 = phonetics2

        if not metrics:
            self.metrics = self._distance_metric[metric_name]
        else:
            self.metrics = metrics

    def distance(self, word1, word2):
        """
        Compute the distance between phonetics codes
        :param word1: first original word
        :param word2: second original word
        :return: distance value
        """
        w1, w2 = self.phonetics1.transform(word1), self.phonetics2.transform(word2)
        if isinstance(self.phonetics1, Soundex) and isinstance(self.phonetics2, Soundex):
            if self.phonetics1.is_delete_first_letter() and self.phonetics2.is_delete_first_letter():
                return self.metrics(w1, w2)
            else:
                return self.metrics(w1[1:], w2[1:])
        return self.metrics(w1, w2)



import re

from abc import ABC, abstractmethod
from ..config import CYRILLIC_SYMBOLS, LATIN_SYMBOLS


class BasePhoneticsAlgorithm(ABC):
    _vowels = ''
    __reduce_regex = re.compile(r'(\w)(\1)+', re.I)
    __cyrillic2latin_table = str.maketrans(CYRILLIC_SYMBOLS, LATIN_SYMBOLS)
    __latin2cyrillic_table = str.maketrans(LATIN_SYMBOLS, CYRILLIC_SYMBOLS)

    def _reduce_seq(self, seq):
        """
        Reduces several repeated symbols in a given string
        :param seq: string
        :return: reduced string
        """
        return self.__reduce_regex.sub(r'\1', seq)

    def _cyrillic2latin(self, seq):
        """
        Converts all cyrillic letters into latin
        :param seq: string
        :return: updated string
        """
        return seq.translate(self.__cyrillic2latin_table)

    def _latin2cyrillic(self, seq):
        """
        Converts all latin letters into cyrillic
        :param seq: string
        :return: updated string
        """
        return seq.translate(self.__latin2cyrillic_table)

    @abstractmethod
    def transform(self, word):
        """
        Converts a given word to phonetic code
        :param word: string
        :return: string code
        """
        return None



from abc import ABC, abstractmethod

from .config import EN_REMOVE_MAP, RU_PHONEMES, RU_REMOVE_MAP, \
    RU_REPLACEMENT_J_MAP, RU_REPLACEMENT_VOWEL_MAP, RU_REPLACEMENT_CONSONANT_MAP, \
    EN_PHONEMES, EE_PHONEMES, FI_PHONEMES, SE_PHONEMES, \
    RU_EGO_OGO_ENDING, RU_IA_ENDING, RU_II_ENDING, EN_METAPHONE_PHONEMES, EN_VOWELS_TO_REMOVE, RU_VOWELS_TO_REMOVE


class RuleSet(ABC):
    @abstractmethod
    def _replacement_phoneme_map(self):
        """
        :return: list of substitution rules
        """
        return []

    def _replace_rules(self, word, rules):
        for replace, result in rules:
            word = replace.sub(result, word)
        return word

    def reduce_phonemes(self, word):
        """
        Transcripts a given word into phonological sequence by language rules
        :param word: string
        :return: modified string
        """
        return self._replace_rules(word, self._replacement_phoneme_map())


class RussianRuleSet(RuleSet):
    """
    Transcription rules for Russian language
    """
    __replacement_j_map = RU_REPLACEMENT_J_MAP
    __replacement_vowel_map = RU_REPLACEMENT_VOWEL_MAP
    __replacement_consonant_map = RU_REPLACEMENT_CONSONANT_MAP
    __remove_map = RU_REMOVE_MAP
    __remove_vowels = RU_VOWELS_TO_REMOVE
    __ia_ending = RU_IA_ENDING
    __ii_ending = RU_II_ENDING
    __ego_ogo_endings = RU_EGO_OGO_ENDING

    def _replacement_phoneme_map(self):
        return RU_PHONEMES

    def replace_consonant_vowels(self, word):
        return self._replace_rules(word, self.__replacement_consonant_map)

    def replace_ego_ogo_ending(self, word):
        return self.__ego_ogo_endings.sub(r'\1в\3', word)

    def replace_ia_ending(self, word):
        return self.__ia_ending.sub('я', word)

    def replace_ii_ending(self, word):
        return self.__ii_ending.sub('и', word)

    def replace_j_and_signs(self, word):
        return self._replace_rules(word, self.__remove_map)

    def replace_j_vowel_phonemes(self, word):
        return self._replace_rules(word, self.__replacement_j_map + self.__replacement_vowel_map)

    def reduce_vowels(self, word):
        return self._replace_rules(word, self.__remove_vowels)


class SwedenRuleSet(RuleSet):
    """
    Transcription rules for Swedish language
    """
    def _replacement_phoneme_map(self):
        return SE_PHONEMES


class EstonianRuleSet(RuleSet):
    """
    Transcription rules for Estonian language
    """
    def _replacement_phoneme_map(self):
        return EE_PHONEMES


class FinnishRuleSet(RuleSet):
    """
    Transcription rules for Finnish language
    """
    def _replacement_phoneme_map(self):
        return FI_PHONEMES


class EnglishRuleSet(RuleSet):
    """
    Transcription rules for English language
    """
    def _replacement_phoneme_map(self):
        return EN_PHONEMES

    __remove_map = EN_REMOVE_MAP

    def remove_empty_sounds(self, word):
        return self._replace_rules(word, self.__remove_map)


class EnglishMetaphoneRuleSet(RuleSet):
    """
    Transcription rules for English language
    """
    def _replacement_phoneme_map(self):
        return EN_METAPHONE_PHONEMES

    __remove_map = EN_REMOVE_MAP
    __remove_vowels = EN_VOWELS_TO_REMOVE

    def remove_empty_sounds(self, word):
        return self._replace_rules(word, self.__remove_map)

    def reduce_vowels(self, word):
        return self._replace_rules(word, self.__remove_vowels)



import re

import pymorphy2

from .base.base import BasePhoneticsAlgorithm
from .config import RU_VOWELS, EN_VOWELS, FI_VOWELS, EE_VOWELS, SE_VOWELS
from .ruleset import EnglishRuleSet, EstonianRuleSet, FinnishRuleSet, RussianRuleSet, SwedenRuleSet


class Soundex(BasePhoneticsAlgorithm):
    """
    Basic class for Soundex algorithm
    """
    _table, _vowels_table = str.maketrans('', ''), str.maketrans('', '')
    _vowels_regex = re.compile(r'(0+)', re.I)

    def __init__(self, delete_first_letter=False, delete_first_coded_letter=False, reduce_word=True,
                 delete_zeros=False, cut_result=False, seq_cutted_len=4, code_vowels=False):
        """
        Initialization of Soundex object
        :param delete_first_letter: remove the first letter from the result code (A169 -> 169)
        :param delete_first_coded_letter: remove the first coded letter from the result code (A0169 -> A169)
        :param reduce_word: reduce repeated characters (A00169 -> A0169)
        :param delete_zeros: remove vowels from the result code
        :param cut_result: cut result core till N symbols
        :param seq_cutted_len: length of the result code
        :param code_vowels: group and code vowels as ABC letters
        """
        self.__delete_first_letter = delete_first_letter
        self.__delete_first_coded_letter = delete_first_coded_letter
        self.__reduce_word = reduce_word
        self.__delete_zeros = delete_zeros
        self.__cut_result = cut_result
        self.__seq_cutted_len = seq_cutted_len
        self.__code_vowels = code_vowels

    def __is_vowel(self, letter):
        return letter in self._vowels

    def __translate_vowels(self, word):
        if self.__code_vowels:
            return word.translate(self._vowels_table)
        else:
            return ''.join('0' if self.__is_vowel(letter) else letter for letter in word)

    def __remove_vowels_and_paired_sounds(self, seq):
        seq = self._vowels_regex.sub('', seq)
        seq = self._reduce_seq(seq)
        return seq

    def _apply_soundex_algorithm(self, word):
        word = word.lower()
        first, last = word[0], word
        last = last.translate(self._table)
        last = self.__translate_vowels(last)
        if self.__reduce_word:
            last = self._reduce_seq(last)
        if self.__delete_zeros:
            last = self.__remove_vowels_and_paired_sounds(last)
        if self.__cut_result:
            last = last[:self.__seq_cutted_len] if len(last) >= self.__seq_cutted_len else last
            last += ('0' * (self.__seq_cutted_len - len(last)))
        if self.__delete_first_coded_letter:
            last = last[1:]
        first_char = '' if self.__delete_first_letter else first.capitalize()
        return first_char + last.upper()

    def get_vowels(self):
        return self._vowels

    def is_delete_first_coded_letter(self):
        return self.__delete_first_coded_letter

    def is_delete_first_letter(self):
        return self.__delete_first_letter

    def transform(self, word):
        return self._apply_soundex_algorithm(word)


class EnglishSoundex(Soundex):
    """
    This version may have differences from original Soundex for English (consonants was splitted in more groups)
    """
    __rule_set = EnglishRuleSet()

    _vowels = EN_VOWELS
    _vowels_table = str.maketrans(_vowels, 'AABBBC')
    _table = str.maketrans('bpfvcksgjqxzdtlmnr', '112233344555667889')

    def _replace_vowels_seq(self, word):
        return self.__rule_set.reduce_phonemes(word)

    def transform(self, word):
        word = self._cyrillic2latin(word)
        word = self.__rule_set.remove_empty_sounds(word)
        return self._apply_soundex_algorithm(word)


class FinnishSoundex(Soundex):
    """
    Soundex for Finnish language
    """
    __rule_set = FinnishRuleSet()

    _vowels = FI_VOWELS
    _vowels_table = str.maketrans(_vowels, 'AAABBBCC')
    _table = str.maketrans('bpfvcszkgqdtlmnrj', '11223334445567789')

    def transform(self, word):
        word = self._cyrillic2latin(word)
        word = self.__rule_set.reduce_phonemes(word)
        return self._apply_soundex_algorithm(word)


class EstonianSoundex(Soundex):
    """
    Soundex for Estonian language
    """
    __rule_set = EstonianRuleSet()

    _vowels = EE_VOWELS
    _vowels_table = str.maketrans(_vowels, 'AAABBBBCC')
    _table = str.maketrans('bpfvcszkgqdtlmnrj', '11223334445567789')

    def transform(self, word):
        word = self._cyrillic2latin(word)
        word = self.__rule_set.reduce_phonemes(word)
        return self._apply_soundex_algorithm(word)


class SwedenSoundex(Soundex):
    """
    Soundex for Sweden language
    """
    __rule_set = SwedenRuleSet()

    _vowels = SE_VOWELS
    _vowels_table = str.maketrans(_vowels, 'AABBBBBCC')
    _table = str.maketrans('bpfvcszkgqdtlmnrj', '11223334445567789')

    def transform(self, word):
        word = self._cyrillic2latin(word)
        if word.endswith('on') and not word.endswith('hon'):
            word = word[:-2] + 'ån'
        word = self.__rule_set.reduce_phonemes(word)
        word = word.replace('sh', 'z')
        word = word.replace('hf', 'x')
        return self._apply_soundex_algorithm(word)


class RussianSoundex(Soundex):
    """
    Soundex for Russian language
    """

    _vowels = RU_VOWELS
    _vowels_table = str.maketrans(_vowels, 'AAAABBBBCC')
    _table = str.maketrans('бпвфгкхдтжшчщзсцлмнр', '11223334455556667889')

    SPEC_ENDING_POSTAGS = {'ADJF', 'NUMB', 'NPRO'}

    def __init__(self, delete_first_letter=False, delete_first_coded_letter=False, reduce_word=True,
                 delete_zeros=False, cut_result=False, seq_cutted_len=4,
                 code_vowels=False, reduce_phonemes=True, replace_ego_ogo_endings=False, use_morph_analysis=False):
        """
        Initialization of Russian Soundex object
        :param delete_first_letter:
        :param delete_first_coded_letter:
        :param reduce_word:
        :param delete_zeros:
        :param code_vowels:
        :param cut_result:
        :param seq_cutted_len:
        :param code_vowels: group and code vowels as ABC letters
        :param reduce_phonemes: simplify sequences of Russian consonants
        :param replace_ego_ogo_endings: replace "-его/-ого" endings with "-ево/-ово"
        :param use_morph_analysis: use morphological analysis for "-его/-ого" replacement
        """
        super(RussianSoundex, self).__init__(delete_first_letter, delete_first_coded_letter, reduce_word,
                                             delete_zeros, cut_result, seq_cutted_len, code_vowels)

        self.reduce_phonemes = reduce_phonemes
        self.rule_set = RussianRuleSet()
        self.use_morph_analysis = use_morph_analysis
        self.replace_ego_ogo_endings = True if self.use_morph_analysis else replace_ego_ogo_endings
        if self.use_morph_analysis:
            self.__moprh = pymorphy2.MorphAnalyzer()

    def __replace_ego_ogo_endings(self, word):
        is_applicable = True
        if self.use_morph_analysis:
            parse = self.__moprh.parse(word)
            is_applicable = parse and any(pos_tag in parse[0].tag for pos_tag in self.SPEC_ENDING_POSTAGS)
        return self.rule_set.replace_ego_ogo_ending(word) if is_applicable else word

    def _replace_vowels_seq(self, word):
        word = self.rule_set.replace_ii_ending(word)
        word = self.rule_set.replace_ia_ending(word)
        return word

    def _reduce_phonemes(self, word):
        word = self.rule_set.replace_consonant_vowels(word)
        word = self.rule_set.replace_j_vowel_phonemes(word)
        word = self.rule_set.reduce_phonemes(word)
        return word

    def transform(self, word):
        """
        Transforms a word into a sequence with coded phonemes
        :param word: string
        :return: Soundex string code
        """
        word = self._latin2cyrillic(word)
        if self.replace_ego_ogo_endings:
            word = self.__replace_ego_ogo_endings(word)
        if self.reduce_phonemes:
            word = self._reduce_phonemes(word)
        word = self.rule_set.replace_j_and_signs(word)
        return self._apply_soundex_algorithm(word)



from setuptools import setup, find_packages
from os.path import join, dirname

setup(
    name='fonetika',
    version='1.5.1',
    url='https://github.com/roddar92/russian_soundex',
    author='Daria Rodionova',
    author_email='drodionova86@gmail.com',
    license='MIT',
    packages=find_packages(),
    description='Phonetics algorithms (Soundex and Metaphone) for russian, english, sweden, finnish and estonian languages',
    long_description=open(join(dirname(__file__), 'README.md')).read(), install_requires=['pymorphy2', 'editdistance'],
    long_description_content_type="text/markdown",
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Natural Language :: Russian',
        'Natural Language :: Finnish',
        'Topic :: Text Processing :: Linguistic'
    ]
)
